searchData={"items":[{"type":"task","title":"mix openapi.dump","doc":"Writes an OpenAPI specification in a JSON file.","ref":"Mix.Tasks.Openapi.Dump.html"},{"type":"task","title":"Synopsis - mix openapi.dump","doc":"mix openapi.dump [options]","ref":"Mix.Tasks.Openapi.Dump.html#module-synopsis"},{"type":"task","title":"Arguments - mix openapi.dump","doc":"* `module` - An Elixir module with a `spec/0` callback returning an OpenAPI\n  specification.","ref":"Mix.Tasks.Openapi.Dump.html#module-arguments"},{"type":"task","title":"Options - mix openapi.dump","doc":"* `-o, --output  ` - The desired output file path. Defaults to `\"openapi.json\"`.\n* `--pretty` - JSON pretty-printing. Defaults to `true`.\n* `--help` - Displays this help.","ref":"Mix.Tasks.Openapi.Dump.html#module-options"},{"type":"function","title":"Mix.Tasks.Openapi.Dump.run/1","doc":"Callback implementation for  Mix.Task.run/1 .","ref":"Mix.Tasks.Openapi.Dump.html#run/1"},{"type":"behaviour","title":"Oaskit","doc":"The main API to work with OpenAPI specifications.\n\nThis module can be used to define a specification module that will then be\nused in your Phoenix router and controllers.","ref":"Oaskit.html"},{"type":"behaviour","title":"Example - Oaskit","doc":"```elixir\ndefmodule MyAppWeb.OpenAPISpec do\n  alias Oaskit.Spec.Paths\n  alias Oaskit.Spec.Server\n  use Oaskit\n\n  @impl true\n  def spec do\n    %{\n      openapi: \"3.1.1\",\n      info: %{title: \"My App API\", version: \"1.0.0\"},\n      servers: [Server.from_config(:my_app, MyAppWeb.Endpoint)],\n      paths: Paths.from_router(MyAppWeb.Router, filter: &String.starts_with?(&1.path, \"/api/\"))\n    }\n  end\nend\n```","ref":"Oaskit.html#module-example"},{"type":"callback","title":"Oaskit.cache/1","doc":"This callback is used to cache the built version of the OpenAPI specification,\nwith JSV schemas turned into validators.\n\nThe callback will be called with `:get` to retrieve a cached build, in which\ncase the callback should return `{:ok, cached}` or `:error`. It will be called\nwith `{:put, value}` to set the cache, in which case it must return `:ok`.\n\nCaching is very important, otherwise the spec will be built for each request\ncalling a controller that uses `ValidateRequest`. An efficient\ndefault implementation using `:persistent_term` is automatically generated.\nOverride this callback if you need more control over the cache.\n\nOn custom implementations there is generally no need to wrap the key in a\ntagged tuple, as it is already a unique tagged tuple.","ref":"Oaskit.html#c:cache/1"},{"type":"type","title":"Oaskit.cache_key/0","doc":"","ref":"Oaskit.html#t:cache_key/0"},{"type":"callback","title":"Oaskit.cache_variant/0","doc":"This function is intended to change cache keys at runtime. The variant is any\nterm used as the last element of a `t:cache_key/0`.\n\nThis is useful if you need to rebuild the OpenAPI specification and its\nvalidators at runtime, when the used schemas or even routes depend on current\napplication state. For instance, if a schema for a given entity is fetched\nregularly from a remote source and changes over time.\n\nThe default implementation returns `nil`.\n\n> #### Stale cache entries are not purged automatically {: .warning}\n>\n> If you return a new variant from this callback, cache entries stored with\n> previous variants in the key are not automatically cleaned. You will need to\n> take care of that. See `c:cache/1` to implement a cache mechanism that you\n> can control.","ref":"Oaskit.html#c:cache_variant/0"},{"type":"function","title":"Oaskit.default_jsv_opts/0","doc":"Default options used for the `JSV.build/2` function when building schemas.","ref":"Oaskit.html#default_jsv_opts/0"},{"type":"callback","title":"Oaskit.jsv_opts/0","doc":"Returns the options that will be passed to `JSV.build/2` when building the\nspec for the implementation module.\n\nThe default implementation delegates to `Oaskit.default_jsv_opts/0`.","ref":"Oaskit.html#c:jsv_opts/0"},{"type":"function","title":"Oaskit.normalize_spec!/1","doc":"Normalizes OpenAPI specification data.\n\nTakes specification data (raw maps or structs) and normalizes it to a\nJSON-compatible version (with binary keys).","ref":"Oaskit.html#normalize_spec!/1"},{"type":"callback","title":"Oaskit.spec/0","doc":"This function should return the OpenAPI specification for your application.\n\nIt can be returned as an `%Elixir.OpenAPI{}` struct, or a bare map with atoms or\nbinary keys (for instance by reading from a JSON file at compile time).\n\nThe returned value will be normalized, any extra data not defined in the\n`Oaskit.Spec...` namespace will be lost.","ref":"Oaskit.html#c:spec/0"},{"type":"function","title":"Oaskit.to_json!/2","doc":"Returns a JSON representation of the given OpenAPI specification module.","ref":"Oaskit.html#to_json!/2"},{"type":"function","title":"Options - Oaskit.to_json!/2","doc":"* `:pretty` - A boolean to control JSON pretty printing.\n* `:validation_error_handler` - A function accepting a `JSV.ValidationError`\n  struct. The JSON generation does not fail on validation error unless you\n  raise from that function.","ref":"Oaskit.html#to_json!/2-options"},{"type":"module","title":"Oaskit.Controller","doc":"Provides macros to define OpenAPI operations directly from controllers.\n\nMacros requires to `use Oaskit.Controller` from your controllers. This\ncan be done wherever `use Phoenix.Controller` is called. With Phoenix, this is\ngenerally in your `MyAppWeb` module, in the `controller` function:\n\n    defmodule MyAppWeb do\n      def controller do\n        quote do\n          use Phoenix.Controller,\n            formats: [:html, :json],\n            layouts: [html: MyAppWeb.Layouts]\n\n          use Oaskit.Controller # <-- Add it there once for all\n\n          use Gettext, backend: MyAppWeb.Gettext\n\n          import Plug.Conn\n\n          # This is alwo where you will plug the validation\n          plug Oaskit.Plugs.ValidateRequest\n\n          unquote(verified_routes())\n        end\n      end\n    end\n\n\nIt can also be useful to define a new `api_controller` function, to separate\ncontrollers that define an HTTP API.\n\nYou would then use that function in your API controllers:\n\n    defmodule MyAppWeb.UserController do\n      use MyAppWeb, :api_controller\n\n      # ...\n    end","ref":"Oaskit.Controller.html"},{"type":"function","title":"Oaskit.Controller.body_params/1","doc":"Returns the validated body from the given `Plug.Conn` struct.","ref":"Oaskit.Controller.html#body_params/1"},{"type":"macro","title":"Oaskit.Controller.operation/2","doc":"Defines an OpenAPI operation for the given Phoenix action (the function listed\nin the router that will handle the conn) that can be validated automatically\nwith the `Oaskit.Plugs.ValidateRequest` plug automatically.\n\nThis macro accepts the function name and a list of options that will define an\n`Oaskit.Spec.Operation`.","ref":"Oaskit.Controller.html#operation/2"},{"type":"macro","title":"Options - Oaskit.Controller.operation/2","doc":"* `:operation_id` - The ID of the operation that is used throughout the\n  validation features. If missing, an id is automatically generated. Operation\n  IDs must be unique.\n* `:tags` - A list of tags (strings) to attach to the operation.\n* `:description` - An optional string to describe the operation in the OpenAPI\n  spec.\n* `:summary` - A short summary of what the operation does.\n* `:parameters` - A keyword list with parameter names as keys and parameter\n  definitions as values. Parameters are query params but also path params. See\n  below for more information.\n* `:request_body` - A map of possible content types and responses definitions.\n  A schema module can be given directly to define a single\n* `:responses` - A map or keyword list where keys are status codes (integers\n  or atoms) and values are responses definitions. See below for responses\n  formats.\n\nPass `false` instead of the options to ignore an action function.","ref":"Oaskit.Controller.html#operation/2-options"},{"type":"macro","title":"Defining parameters - Oaskit.Controller.operation/2","doc":"Parameters are organized by their name and their `:in` option. Two parameters\nwith the same key can coexist if their `:in` option is different. The `:query`\nand `:path` values for `:in` are currently supported.\n\nParameters support the following options:\n\n* `:in` - Either `:path` or `:query`. Required.\n* `:schema` - A JSON schema or Module name exporting a `schema/0` function.\n* `:required` - A boolean, defaults to `true` for `:path` params, `false`\n  otherwise.\n* `:examples` - A list of examples.\n\nParameters are stored into `conn.private.oaskit.path_params` and\n`conn.private.oaskit.query_params`. They do not override the `params`\nargument passed to your phoenix action function. Those original `params` are\nstill the ones as decoded by phoenix.","ref":"Oaskit.Controller.html#operation/2-defining-parameters"},{"type":"macro","title":"Parameters example - Oaskit.Controller.operation/2","doc":"# Imaginary GET /api/users/:organization route\n\n    operation :list_users,\n      operation_id: \"ListUsers\",\n      parameters: [\n        organization: [in: :path,  required: true,  schema: %{type: :string}],\n        page:         [in: :query, required: false, schema: %{type: :integer, minimum: 1}],\n        per_page:     [in: :query, required: false, schema: %{type: :integer, minimum: 1}]\n      ],\n      # ...\n\n    def list_users(conn, _params) do\n      page = query_param(conn, :page)\n      per_page = query_param(conn, :per_page)\n      do_something_with(conn, page, per_page)\n    end","ref":"Oaskit.Controller.html#operation/2-parameters-example"},{"type":"macro","title":"Defining the request body - Oaskit.Controller.operation/2","doc":"Request bodies can be defined in two ways: Either by providing a mapping of\ncontent-type to media type objects, or with a shortcut by providing only a\nschema for an unique `\"application/json\"` content-type.\n\nThe body can be retrieved in `conn.oaskit.private.body_params`.\n\nOptions supported with a generic definition, for each content type:\n\n* `:content` - A map of content-type to bodies definitions. Content-types\n  should be strings.\n* `:required` - A boolean. When `false`, the body can be missing and will not\n  be validated. In that case, `conn.oaskit.private.body_params` will be\n  `nil`. The default value is `false`.\n\nWhen using the shortcut, a single atom or 2-tuple is expected.\n\n* Supported atoms are `true` (a JSON schema that accepts anything), `false` (a\n  JSON schema that rejects everything) or a module name. The module must\n  export a `schema/0` function that returns a JSON schema.\n* When passing a tuple, the first element is a schema (boolean or module), but\n  a direct JSON schema map (like `%{type: :object, ...}`) is also accepted.\n  The second tuple element is a list of options for the response body object.\n\n**Important**, when using the shortcut, we chose to automatically define the\n`:required` option of the media type object to `true`.","ref":"Oaskit.Controller.html#operation/2-defining-the-request-body"},{"type":"macro","title":"Request body examples - Oaskit.Controller.operation/2","doc":"A short form using a module schema:\n\n    operation :create_user,\n      operation_id: \"CreateUser\",\n      request_body: UserSchema,\n      # ...\n\n    def create_user(conn, _params) do\n      case Users.create_user(conn.private.oaskit.body_params) do\n        # ...\n      end\n    end\n\nThe operation definition above is equivalent to this:\n\n    operation :create_user,\n      operation_id: \"CreateUser\",\n      request_body: [\n        content: %{\"application/json\" => %{schema: CreateUserPayload}},\n        required: true\n      ],\n      # ...\n\nTo make the body non-required in the short form, use the tuple version:\n\n    operation :create_user,\n      operation_id: \"CreateUser\",\n      request_body: {UserSchema, required: false},\n      # ...\n\nMultiple content-types can be supported. Content-types with wildcards will be\ntried last by `Plug.Parsers`, as well as oaskit when choosing\nthe schema for validation.\n\n    operation :create_user,\n      request_body: [\n        content: %{\n          \"application/x-www-form-urlencoded\" => %{schema: CreateUserPayload},\n          \"application/json\" => %{schema: CreateUserPayload},\n          \"*/*\" => %{schema: %{type: :string}}\n        }\n      ]","ref":"Oaskit.Controller.html#operation/2-request-body-examples"},{"type":"macro","title":"Defining responses - Oaskit.Controller.operation/2","doc":"Responses are defined by a mapping of HTTP statuses to response objects.\n\n* HTTP statuses can be given as integers (`200`, `404`, _etc._) or atoms\n  supported by `Plug.Conn.Status` like `:ok`, `:not_found`, _etc_.\n* `:default` can be given instead of a status to define the default option\n  supported by the OpenAPI speficication. This is often used to define a\n  generic error response.\n\nResponse objects accept the following options:\n\n* `:description` - This is mandatory for responses.\n* `:headers` and `:links` - This is not used by the validation mechanisms of\n  this library, but is useful to be defined in the OpenAPI specification JSON\n  document.\n* `:content` - A mapping of content-type to media type objects, exactly as in\n  the request bodies definitions.\n\nFinally, the response for each status can also be defined with a shortcut, by\nusing a single schema that will be associated to the `\"application/json\"`\ncontent-type. The mandatory description can be provided when using the tuple\nshortcut, or will otherwise being pulled from the schema `description`\nkeyword.","ref":"Oaskit.Controller.html#operation/2-defining-responses"},{"type":"macro","title":"Reponse examples - Oaskit.Controller.operation/2","doc":"A first example using the atom statuses, and a shortcut for the full response\ndefinition:\n\n    operation :list_users,\n      operation_id: \"ListUsers\",\n      responses: [ok: UsersListPage]\n\nThe definition above is equivalent to the following:\n\n    operation :list_users,\n      operation_id: \"ListUsers\",\n      responses: %{\n        200 => [\n          description: UsersListPage.schema().description,\n          content: %{\n            \"application/json\" => %{schema: UsersListPage}\n          }\n        ]\n      }\n\nThe description can be overriden when using the shortcut:\n\n    operation :list_users,\n      operation_id: \"ListUsers\",\n      responses: [ok: {UsersListPage, description: \"A page of users\"}]\n\nMultiple status codes are generally expected. The shortcut can be used in only\na part of them.\n\n    operation :list_users,\n      operation_id: \"ListUsers\",\n      responses: [\n        ok: UsersListPage,\n        not_found: {GenericErrorSchema, description: \"not found generic response\"},\n        forbidden: {%{type: :array}, description: \"missing-role messages\"},\n        internal_server_error: [\n          description: \"Error with stacktrace\",\n          content: %{\n            \"application/json\" => [\n              schema: %{type: :array, items: %{type: :string, description: \"trace item\"}}\n            ],\n            \"text/plain\" => [schema: true]\n          }\n        ]\n      ]\n\nOf course, mixing all styles together is discouraged for readability.","ref":"Oaskit.Controller.html#operation/2-reponse-examples"},{"type":"macro","title":"Ignore operations - Oaskit.Controller.operation/2","doc":"","ref":"Oaskit.Controller.html#operation/2-ignore-operations"},{"type":"macro","title":"Oaskit.Controller.parameter/2","doc":"Defines a parameter for all operations defined _later_ in the module body with\nthe `operation/2` macro.\n\nTakes the same options as the `:parameters` option items from that macro.\n\nIf an operation also directly defines a parameter with the same `name` and\n`:in` option, it will take precedence and the parameter defined with\n`parameter/2` will be ignored.","ref":"Oaskit.Controller.html#parameter/2"},{"type":"macro","title":"Example - Oaskit.Controller.parameter/2","doc":"In the following example, the second operation defines its own version of the\n`per_page` parameter to limit the number of users returned in a single page.\n\n    # This macro can be called multiple times\n    parameter :slug, in: :path, schema: %{type: :string, pattern: \"[0-9a-z-]+\"}\n    parameter :page, in: :query, schema: %{type: :integer, minimum: 1}\n    parameter :per_page, in: :query, schema: %{type: :integer, minimum: 1, maximum: 100}\n\n    operation :list_users, operation_id: \"ListUsers\", responses: [ok: UsersListPage]\n\n    def list_users(conn, params) do\n      # ...\n    end\n\n    operation :list_users_deep,\n      operation_id: \"ListUsersDeep\",\n      parameters: [\n        per_page: [in: :query, schema: %{type: :integer, minimum: 1, maximum: 20}]\n      ],\n      responses: [\n        ok:\n          {DeepUsersListPage,\n           description: \"Returns users with all associated organization and blog posts data\"}\n      ]\n\n    def list_users_deep(conn, params) do\n      # ...\n    end","ref":"Oaskit.Controller.html#parameter/2-example"},{"type":"function","title":"Oaskit.Controller.path_param/3","doc":"Accepts a `Plug.Conn` struct, a parameter name (as atom) and a default value.\n\nReturns the validated parameter from `conn.oaskit.private.path_params` if\nfound, or the default value.","ref":"Oaskit.Controller.html#path_param/3"},{"type":"function","title":"Oaskit.Controller.query_param/3","doc":"Accepts a `Plug.Conn` struct, a parameter name (as atom) and a default value.\n\nReturns the validated parameter from `conn.oaskit.private.query_params` if\nfound, or the default value.","ref":"Oaskit.Controller.html#query_param/3"},{"type":"macro","title":"Oaskit.Controller.tags/1","doc":"Defines tags for all operations defined _later_ in the module body with the\n`operation/2` macro.\n\nIf an operation also directly defines tags, they will be merged.","ref":"Oaskit.Controller.html#tags/1"},{"type":"macro","title":"Example - Oaskit.Controller.tags/1","doc":"# This macro can be called multiple times\n    tags [\"users\", \"v1\"]\n    tags [\"other-tag\"]\n\n    operation :list_users,\n      operation_id: \"ListUsers\",\n      responses: [ok: UsersListPage]\n\n    def list_users(conn, params) do\n      # ...\n    end\n\n    operation :list_users_deep,\n      operation_id: \"ListUsersDeep\",\n      tags: [\"slow\"],\n      responses: [ok: DeepUsersListPage]\n\n    def list_users_deep(conn, params) do\n      # ...\n    end","ref":"Oaskit.Controller.html#tags/1-example"},{"type":"macro","title":"Oaskit.Controller.use_operation/3","doc":"This macro allows Oaskit to validate request bodies, query and path\nparameters (and responses in tests) when an OpenAPI specification is not\ndefined with the `operation/2` macro but rather provided directly in an\nexternal spec document.\n\nFor instance with the following spec module:\n\n    defmodule MyAppWeb.ExternalAPISpec do\n      use Oaskit\n      @api_spec JSON.decode!(File.read!(\"priv/api/spec.json\"))\n\n      @impl true\n      def spec, do: @api_spec\n    end\n\nGiven the `spec.json` file decribes an operation whose `operationId` is\n`\"ListUsers\"`, then the request/response validation can be enabled like this:\n\n    use_operation :list_users, \"ListUsers\"\n\n    def list_users(conn, params) do\n      # ...\n    end\n\n> #### Parameter names always create atoms {: .warning}\n>\n> Query and path parameters defined in OpenAPI specifications always define\n> the corresponding atoms, even if that specification is read from a JSON\n> file, or defined manually in code with string keys.\n>\n> For that reason it is ill advised to use specs generated dynamically at\n> runtime without validating their content.","ref":"Oaskit.Controller.html#use_operation/3"},{"type":"module","title":"Oaskit.Plugs.SpecProvider","doc":"A plug to associate an OpenAPI specification with a group of routes in a\nrouter or a controller.\n\nIt takes a `:spec` option with the name of a module implementing the\n`Oaskit` behaviour.\n\nIt will generally be used from a `Phoenix.Router` implementation:\n\n    defmodule MyAppWeb.Router do\n      use Phoenix.Router\n\n      # The provider should be called in a pipeline.\n      pipeline :api do\n        plug Oaskit.Plugs.SpecProvider, spec: MyAppWeb.ApiSpec\n      end\n\n      scope \"/api\", MyAppWeb.Api do\n        # Then that pipeline can be used in one or\n        # more scopes.\n        pipe_through :api\n\n        # Controllers used in such scopes can now use\n        # the `Oaskit.Plugs.ValidateRequest` plug.\n        get \"/hello\", HelloController, :hello\n      end\n    end\n\n>","ref":"Oaskit.Plugs.SpecProvider.html"},{"type":"module","title":"Why do we need this? - Oaskit.Plugs.SpecProvider","doc":">\n> Why not directly pass the spec module to `Oaskit.Plugs.ValidateRequest`?\n>\n> Because we may want to attach a controller action and its operation ID to\n> multiple API specifications.\n>\n> For that reason, specs are attached to routes using a pipeline, and not to\n> controllers. This is why this plug is used in router modules, while the\n> `Oaskit.Plugs.ValidateRequest` plug will take whatever spec was given in the\n> conn and fetch the operation ID from there.","ref":"Oaskit.Plugs.SpecProvider.html#module-why-do-we-need-this"},{"type":"function","title":"Oaskit.Plugs.SpecProvider.call/2","doc":"Callback implementation for  Plug.call/2 .","ref":"Oaskit.Plugs.SpecProvider.html#call/2"},{"type":"function","title":"Oaskit.Plugs.SpecProvider.fetch_spec_module!/1","doc":"Returns the spec module from a conn if the conn went through this plug, raises\nan error otherwise.","ref":"Oaskit.Plugs.SpecProvider.html#fetch_spec_module!/1"},{"type":"function","title":"Oaskit.Plugs.SpecProvider.init/1","doc":"Callback implementation for  Plug.init/1 .","ref":"Oaskit.Plugs.SpecProvider.html#init/1"},{"type":"module","title":"Oaskit.Plugs.ValidateRequest","doc":"This plug will match incoming requests to operations defined with the\n`Oaskit.Controller.operation/2` or\n`Oaskit.Controller.use_operation/2` macros and retrieved from a\nprovided OpenAPI Specification module.","ref":"Oaskit.Plugs.ValidateRequest.html"},{"type":"module","title":"Pluggin' in - Oaskit.Plugs.ValidateRequest","doc":"To use this plug in a controller, the\n`Oaskit.Plugs.SpecProvider` plug must be used in the router for\nthe corresponding routes.\n\n    defmodule MyAppWeb.Router do\n      use Phoenix.Router\n\n      pipeline :api do\n        plug Oaskit.Plugs.SpecProvider, spec: MyAppWeb.OpenAPISpec\n      end\n\n      scope \"/api\" do\n        pipe_through :api\n\n        scope \"/users\", MyAppWeb do\n          get \"/\", UserController, :list_users\n        end\n      end\n    end\n\nThis plug must then be used in controllers. It is possible to call the plug in\nevery controller where you want validation, or to define it globally in your\n`MyAppWeb` module.\n\nWhile you can directly patch the `MyAppWeb.controller` function if all your\ncontrollers belong to the HTTP API, we suggest to create a new\n`api_controller` function in your `MyAppWeb` module.\n\nDuplicate the `controller` function and add this plug and also `use\nOaskit.Controller`.\n\n    defmodule MyAppWeb do\n      def controller do\n        # ...\n      end\n\n      def api_controller do\n        quote do\n          use Phoenix.Controller, formats: [:json], layouts: []\n\n          # Use the controller helpers to define operations\n          use Oaskit.Controller\n\n          use Gettext, backend: MyAppWeb.Gettext\n          import Plug.Conn\n\n          # Use the plug here. This has to be after `use Phoenix.Controller`.\n          plug Oaskit.Plugs.ValidateRequest\n\n          unquote(verified_routes())\n        end\n      end\n    end\n\nFinally, request body validation will only work if the body is fetched from\nthe conn. This is generally done by the `Plug.Parsers` plug. It can also be\ndone by a custom plug if you are implementing an API that is working with\nplaintext or custom formats.","ref":"Oaskit.Plugs.ValidateRequest.html#module-pluggin-in"},{"type":"module","title":"Request validation - Oaskit.Plugs.ValidateRequest","doc":"Requests will be validated according to the request body schemas and parameter\nschemas defined in operations. The data will also be cast to the expected\ntypes:\n\n* Parameters whose schemas define a `type` of `boolean` or `integer` will be\n  cast to that type. Arrays of such types are supported as well.\n* Parameters with type `string` and a `format` supported by `JSV` will be also\n  cast according to that format. Output values for formats are described in\n  the [JSV\n  documentation](https://hexdocs.pm/jsv/validation-basics.html#formats). This\n  includes, URI, Date, DateTime and Duration.\n* Request bodies are cast to their given schema too. When using raw schemas\n  defined as maps, the main changes to the data is the cast of formats, just\n  as for parameters. When schemas are defined using module names, and when\n  those modules' structs are created with `JSV.defschema/1`, the request\n  bodies will be cast to those structs.\n\nRequest bodies will be validated according to the expected operation\ncontent-types, and the actual content-type of the request.","ref":"Oaskit.Plugs.ValidateRequest.html#module-request-validation"},{"type":"module","title":"Options - Oaskit.Plugs.ValidateRequest","doc":"* `:query_reader_opts` - If a Plug.Conn struct enters this plug without its\n  query parameters being fetched, this plug will fetch them automatically\n  using `Conn.fetch_query_params(conn, query_reader_opts)`. The default value\n  is `[length: 1_000_000, validate_utf8: true]`.\n* `:error_handler` - A module or `{module, argument}` tuple. The error handler\n  must implement the `Oaskit.ErrorHandler` behaviour. It will be\n  called on validation errors. Defaults to `Oaskit.ErrorHandler.Default`.\n* `:pretty_errors` - A boolean to control pretty printing of JSON errors\n  payload in error handlers. Defaults to `true` when `Mix.env() != :prod`,\n  defaults to `false` otherwise.\n* `:html_errors` - A boolean to control whether the default error handler is\n  allowed to return HTML errors when the request accepts HTML. This is useful\n  to quickly read errors when opening an url directly from the browser.\n  Defaults to `true`.\n* Unknown options are collected and passed to the error handler.","ref":"Oaskit.Plugs.ValidateRequest.html#module-options"},{"type":"module","title":"Non-required bodies - Oaskit.Plugs.ValidateRequest","doc":"A request body is considered empty if `\"\"`, `nil` or an empty map (`%{}`). In\nthat case, if the operation defines the request body with `required: false`\n(which is the default value!), the body validation will be skipped.\n\nThe empty map is a special case because `Plug.Parsers` implementations cannot\nreturn anything else than a map. If a client sends an HTTP request with an\n`\"application/json\"` content-type but no body, the JSON parser in the Plug\nlibrary will still return an empty map.\n\nTo avoid problems, always define request bodies as required if you can. This\nis made automatically when using the \"definition shortcuts\" described in\n`Oaskit.Controller.operation/2`.","ref":"Oaskit.Plugs.ValidateRequest.html#module-non-required-bodies"},{"type":"module","title":"Error handling - Oaskit.Plugs.ValidateRequest","doc":"Validation can fail at various steps:\n\n* Parameters validation\n* Content-type matching\n* Body validation\n\nOn failure, the validation stops immediately. If a parameter is invalid, the\nbody is not validated and the error handler is called with a single category\nof errors. In the case of parameters, multiple errors can be passed to the\nhandler if multiple parameters are invalid or missing.\n\nCustom error handlers must implement the `Oaskit.ErrorHandler` behaviour and\nbe ready to accept all error reasons that this plug can generate. Such reasons\nare described in the `t:Oaskit.ErrorHandler.reason/0` type.\n\nThe 3rd argment passed to the `c:Oaskit.ErrorHandler.handle_error/3` depends\non the `:error_handler` function. When defined as a module, the argument is\nall the other options given to this plug:\n\n    plug Oaskit.Plugs.ValidateRequest,\n      error_handler: MyErrorHandler,\n      pretty_errors: true,\n      custom_opt: \"foo\"\n\nThis will allow the handler to receive `:pretty_errors`, `:custom_opt` and\nother options with default values.\n\nWhen passing a tuple, the second element will be passed as-is:\n\n    plug Oaskit.Plugs.ValidateRequest,\n      error_handler: {MyErrorHandler, \"foo\"}\n\nIn the example above, the error handler will only receive `\"foo\"` as the 3rd\nargument of `c:Oaskit.ErrorHandler.handle_error/3`.","ref":"Oaskit.Plugs.ValidateRequest.html#module-error-handling"},{"type":"module","title":"Oaskit.SpecController","doc":"Controller for serving OpenAPI specifications in JSON format or Redoc UI.","ref":"Oaskit.SpecController.html"},{"type":"module","title":"Serving the JSON spec - Oaskit.SpecController","doc":"Add a route in your router by passing the spec module to this controller:\n\n    get \"/openapi.json\", Oaskit.SpecController, spec: MyAppWeb.ApiSpec\n\nAlternatively, you can automatically serve the spec that provided with the\nspec provider plug by replacing the options with `:show`. The route needs to\nbe scoped with the appropriate pipeline:\n\n    defmodule MyAppWeb.Router do\n      use Phoenix.Router\n\n      pipeline :api do\n        plug Oaskit.Plugs.SpecProvider, spec: MyAppWeb.ApiSpec\n      end\n\n      scope \"/\" do\n        pipe_through :api\n\n        get \"/openapi.json\", Oaskit.SpecController, :show\n      end\n    end\n\nThis will serve your OpenAPI specification at `/openapi.json`. Add `?pretty=1`\nin the URL for easier debugging.\n\nYou can also replace `:show` by `spec: MyAppWeb.ApiSpec` when calling the\ncontroller\n\nThe controller works with any spec module that implements the `Oaskit`\nbehavior. Note that if your spec is obtained from a static file, you don't\nreally need that controller and can just serve that file.","ref":"Oaskit.SpecController.html#module-serving-the-json-spec"},{"type":"module","title":"Showing Redoc - Oaskit.SpecController","doc":"Redoc is a simple documentation generator that will display an OpenAPI\nspecification in a nice way. It does not allow to test the routes though.\n\nPass the absolute URL path to the controller with the `:redoc` option, and\noptionally a redoc configuration object:\n\n    get \"/redoc\", Oaskit.SpecController, redoc: \"/url/to/openapi.json\"\n    get \"/redoc\", Oaskit.SpecController,\n      redoc: \"/generated/openapi.json\",\n      redoc_config: %{\n        \"minCharacterLengthToInitSearch\" => 1,\n        \"hideDownloadButtons\" => true\n      }","ref":"Oaskit.SpecController.html#module-showing-redoc"},{"type":"function","title":"Oaskit.SpecController.call/2","doc":"Callback implementation for  Plug.call/2 .","ref":"Oaskit.SpecController.html#call/2"},{"type":"function","title":"Oaskit.SpecController.init/1","doc":"Callback implementation for  Plug.init/1 .","ref":"Oaskit.SpecController.html#init/1"},{"type":"function","title":"Oaskit.SpecController.serve_spec/2","doc":"","ref":"Oaskit.SpecController.html#serve_spec/2"},{"type":"behaviour","title":"Oaskit.ErrorHandler","doc":"A behaviour for validation errors handlers.","ref":"Oaskit.ErrorHandler.html"},{"type":"callback","title":"Oaskit.ErrorHandler.handle_error/3","doc":"Accepts the Plug.Conn struct, an error reason and the options passed to the\n`Oaskit.Plugs.ValidateRequest` plug.\n\nThis function is called when request validation fails and an error must be\nreturned to the remote client. This means that function _must_ send a\nresponse.\n\nResponses can be sent just as in Phoenix controllers, using\n`Plug.Conn.send_resp/3`, `Phoenix.Controller.json/2`,\n`Phoenix.Controller.text/2`, _etc._\n\nThe `arg` argument is the options given to `Oaskit.Plugs.ValidateRequest`. See\nthis module documentation for more information.","ref":"Oaskit.ErrorHandler.html#c:handle_error/3"},{"type":"type","title":"Oaskit.ErrorHandler.reason/0","doc":"","ref":"Oaskit.ErrorHandler.html#t:reason/0"},{"type":"module","title":"Oaskit.ErrorHandler.Default","doc":"The default error handler for validation errors.\n\nReturns various HTTP error codes depending on the validation failure:\n\n* `422`\n  _Unprocessable Entity_ - When JSON\n  schema errors when validating request bodies.\n* `415`\n  _Unsupported Media Type_ -\n  When matching the request content-type to the accepted content-types.\n* `400`\n  _Bad Request_ - When\n  validating query and path parameters.","ref":"Oaskit.ErrorHandler.Default.html"},{"type":"module","title":"Error formatting - Oaskit.ErrorHandler.Default","doc":"For request specifically accepting HTML (`\"html\"` is found in the Accept\nheader), this handler will present errors in an HTML document with basic\nstyles.\n\nIn any other case, a JSON representation of errors is returned with the\n`\"application/json\"` content type.\n\nDisable HTML entirely with the `html_errors: false` option when using the\n`Oaskit.Plugs.ValidateRequest` plug.","ref":"Oaskit.ErrorHandler.Default.html#module-error-formatting"},{"type":"function","title":"Oaskit.ErrorHandler.Default.error_response_schema/0","doc":"Returns a module-based JSON schema for json responses returned by this module\nwhen used as the handler of validation errors.","ref":"Oaskit.ErrorHandler.Default.html#error_response_schema/0"},{"type":"function","title":"Example - Oaskit.ErrorHandler.Default.error_response_schema/0","doc":"operation :my_action,\n      request_body: UserCreationParams,\n      responses: [\n        ok: User,\n        default: Oaskit.ErrorHandler.Default.error_response_schema()\n      ]\n\n    def my_action(conn, params) do\n      # ...\n    end","ref":"Oaskit.ErrorHandler.Default.html#error_response_schema/0-example"},{"type":"function","title":"Oaskit.ErrorHandler.Default.handle_error/3","doc":"Callback implementation for  Oaskit.ErrorHandler.handle_error/3 .","ref":"Oaskit.ErrorHandler.Default.html#handle_error/3"},{"type":"module","title":"Oaskit.Test","doc":"Provides the `valid_response/3` test helper to validate API responses in your\nExUnit tests.","ref":"Oaskit.Test.html"},{"type":"function","title":"Oaskit.Test.valid_response/3","doc":"Validates that the given conn bears a response that is valid with regard to\nthe OpenAPI operation that served it, and returns the response body.\n\nResponses returned with a JSON based content-type like `\"application/json\"` or\n`\"application/vnd.api+json\"` will be decoded.\n\nIt is encouraged to wrap this function with a custom helper, typically in your\n`MyAppWeb.ConnCase` test helper:\n\n    defmodule MyAppWeb.ConnCase do\n\n      # ...\n\n      # You can wrap the helper function this way so you do not have to pass\n      # The spec module in every call:\n      def valid_response(conn, status) do\n        Oaskit.Test.valid_response(MyAppWeb.OpenAPISpec, conn, status)\n      end\n    end\n\n\nThen use the helper in your tests:\n\n    test \"should return the user info\", %{conn: conn} do\n      %{id: id} = user_fixture(username: \"joe\", roles: [\"admin\"])\n      conn = get(conn, ~p\"/api/users/#{id}\")\n\n      assert %{\n        \"username\" => \"joe\",\n        \"roles\" => [\"admin\"]\n      } = valid_response(conn, 200)\n    end","ref":"Oaskit.Test.html#valid_response/3"},{"type":"module","title":"Oaskit.Spec.Callback","doc":"Representation of the [ Callback Object](https://spec.openapis.org/oas/v3.1.1.html#callback-object) in OpenAPI Specification.","ref":"Oaskit.Spec.Callback.html"},{"type":"function","title":"Oaskit.Spec.Callback.schema/0","doc":"","ref":"Oaskit.Spec.Callback.html#schema/0"},{"type":"module","title":"Oaskit.Spec.Components","doc":"Representation of the [ Components Object](https://spec.openapis.org/oas/v3.1.1.html#components-object) in OpenAPI Specification.","ref":"Oaskit.Spec.Components.html"},{"type":"function","title":"Oaskit.Spec.Components.schema/0","doc":"","ref":"Oaskit.Spec.Components.html#schema/0"},{"type":"module","title":"Oaskit.Spec.Contact","doc":"Representation of the [ Contact Object](https://spec.openapis.org/oas/v3.1.1.html#contact-object) in OpenAPI Specification.","ref":"Oaskit.Spec.Contact.html"},{"type":"function","title":"Oaskit.Spec.Contact.schema/0","doc":"","ref":"Oaskit.Spec.Contact.html#schema/0"},{"type":"module","title":"Oaskit.Spec.Discriminator","doc":"Representation of the [ Discriminator Object](https://spec.openapis.org/oas/v3.1.1.html#discriminator-object) in OpenAPI Specification.","ref":"Oaskit.Spec.Discriminator.html"},{"type":"function","title":"Oaskit.Spec.Discriminator.schema/0","doc":"","ref":"Oaskit.Spec.Discriminator.html#schema/0"},{"type":"module","title":"Oaskit.Spec.Encoding","doc":"Representation of the [ Encoding Object](https://spec.openapis.org/oas/v3.1.1.html#encoding-object) in OpenAPI Specification.","ref":"Oaskit.Spec.Encoding.html"},{"type":"function","title":"Oaskit.Spec.Encoding.schema/0","doc":"","ref":"Oaskit.Spec.Encoding.html#schema/0"},{"type":"module","title":"Oaskit.Spec.Example","doc":"Representation of the [ Example Object](https://spec.openapis.org/oas/v3.1.1.html#example-object) in OpenAPI Specification.","ref":"Oaskit.Spec.Example.html"},{"type":"function","title":"Oaskit.Spec.Example.schema/0","doc":"","ref":"Oaskit.Spec.Example.html#schema/0"},{"type":"module","title":"Oaskit.Spec.ExternalDocumentation","doc":"Representation of the [ External Documentation Object](https://spec.openapis.org/oas/v3.1.1.html#external-documentation-object) in OpenAPI Specification.","ref":"Oaskit.Spec.ExternalDocumentation.html"},{"type":"function","title":"Oaskit.Spec.ExternalDocumentation.schema/0","doc":"","ref":"Oaskit.Spec.ExternalDocumentation.html#schema/0"},{"type":"module","title":"Oaskit.Spec.Header","doc":"Representation of the [ Header Object](https://spec.openapis.org/oas/v3.1.1.html#header-object) in OpenAPI Specification.","ref":"Oaskit.Spec.Header.html"},{"type":"function","title":"Oaskit.Spec.Header.schema/0","doc":"","ref":"Oaskit.Spec.Header.html#schema/0"},{"type":"module","title":"Oaskit.Spec.Info","doc":"Representation of the [ Info Object](https://spec.openapis.org/oas/v3.1.1.html#info-object) in OpenAPI Specification.","ref":"Oaskit.Spec.Info.html"},{"type":"function","title":"Oaskit.Spec.Info.schema/0","doc":"","ref":"Oaskit.Spec.Info.html#schema/0"},{"type":"module","title":"Oaskit.Spec.License","doc":"Representation of the [ License Object](https://spec.openapis.org/oas/v3.1.1.html#license-object) in OpenAPI Specification.","ref":"Oaskit.Spec.License.html"},{"type":"function","title":"Oaskit.Spec.License.schema/0","doc":"","ref":"Oaskit.Spec.License.html#schema/0"},{"type":"module","title":"Oaskit.Spec.Link","doc":"Representation of the [ Link Object](https://spec.openapis.org/oas/v3.1.1.html#link-object) in OpenAPI Specification.","ref":"Oaskit.Spec.Link.html"},{"type":"function","title":"Oaskit.Spec.Link.schema/0","doc":"","ref":"Oaskit.Spec.Link.html#schema/0"},{"type":"module","title":"Oaskit.Spec.MediaType","doc":"Representation of the [ Media Type Object](https://spec.openapis.org/oas/v3.1.1.html#media-type-object) in OpenAPI Specification.","ref":"Oaskit.Spec.MediaType.html"},{"type":"function","title":"Oaskit.Spec.MediaType.from_controller!/1","doc":"","ref":"Oaskit.Spec.MediaType.html#from_controller!/1"},{"type":"function","title":"Oaskit.Spec.MediaType.schema/0","doc":"","ref":"Oaskit.Spec.MediaType.html#schema/0"},{"type":"module","title":"Oaskit.Spec.OAuthFlow","doc":"Representation of the [ O Auth Flow Object](https://spec.openapis.org/oas/v3.1.1.html#o-auth-flow-object) in OpenAPI Specification.","ref":"Oaskit.Spec.OAuthFlow.html"},{"type":"function","title":"Oaskit.Spec.OAuthFlow.schema/0","doc":"","ref":"Oaskit.Spec.OAuthFlow.html#schema/0"},{"type":"module","title":"Oaskit.Spec.OAuthFlows","doc":"Representation of the [ O Auth Flows Object](https://spec.openapis.org/oas/v3.1.1.html#o-auth-flows-object) in OpenAPI Specification.","ref":"Oaskit.Spec.OAuthFlows.html"},{"type":"function","title":"Oaskit.Spec.OAuthFlows.schema/0","doc":"","ref":"Oaskit.Spec.OAuthFlows.html#schema/0"},{"type":"module","title":"Oaskit.Spec.OpenAPI","doc":"Representation of the [ Open Api Object](https://spec.openapis.org/oas/v3.1.1.html#open-api-object) in OpenAPI Specification.","ref":"Oaskit.Spec.OpenAPI.html"},{"type":"function","title":"Oaskit.Spec.OpenAPI.schema/0","doc":"","ref":"Oaskit.Spec.OpenAPI.html#schema/0"},{"type":"module","title":"Oaskit.Spec.Operation","doc":"Representation of the [ Operation Object](https://spec.openapis.org/oas/v3.1.1.html#operation-object) in OpenAPI Specification.","ref":"Oaskit.Spec.Operation.html"},{"type":"function","title":"Oaskit.Spec.Operation.from_controller!/2","doc":"","ref":"Oaskit.Spec.Operation.html#from_controller!/2"},{"type":"function","title":"Oaskit.Spec.Operation.schema/0","doc":"","ref":"Oaskit.Spec.Operation.html#schema/0"},{"type":"module","title":"Oaskit.Spec.Parameter","doc":"Representation of the [ Parameter Object](https://spec.openapis.org/oas/v3.1.1.html#parameter-object) in OpenAPI Specification.","ref":"Oaskit.Spec.Parameter.html"},{"type":"function","title":"Oaskit.Spec.Parameter.from_controller!/2","doc":"","ref":"Oaskit.Spec.Parameter.html#from_controller!/2"},{"type":"function","title":"Oaskit.Spec.Parameter.schema/0","doc":"","ref":"Oaskit.Spec.Parameter.html#schema/0"},{"type":"module","title":"Oaskit.Spec.PathItem","doc":"Representation of the [ Path Item Object](https://spec.openapis.org/oas/v3.1.1.html#path-item-object) in OpenAPI Specification.","ref":"Oaskit.Spec.PathItem.html"},{"type":"function","title":"Oaskit.Spec.PathItem.schema/0","doc":"","ref":"Oaskit.Spec.PathItem.html#schema/0"},{"type":"function","title":"Oaskit.Spec.PathItem.verbs/0","doc":"","ref":"Oaskit.Spec.PathItem.html#verbs/0"},{"type":"module","title":"Oaskit.Spec.Paths","doc":"Representation of the [ Paths Object](https://spec.openapis.org/oas/v3.1.1.html#paths-object) in OpenAPI Specification.","ref":"Oaskit.Spec.Paths.html"},{"type":"function","title":"Oaskit.Spec.Paths.from_router/2","doc":"Accepts a Phoenix router module and returns paths that point to a controller\nwith operations defined for its actions.","ref":"Oaskit.Spec.Paths.html#from_router/2"},{"type":"function","title":"Options - Oaskit.Spec.Paths.from_router/2","doc":"* `:filter` - A predicate function to limit routes defined in your OpenAPI\n  specification. This predicate is not called for every route, only on routes\n  that define an operation.","ref":"Oaskit.Spec.Paths.html#from_router/2-options"},{"type":"function","title":"Oaskit.Spec.Paths.from_routes/2","doc":"Same as `from_router/2` but directly accepts the return value of\n`router_module.__routes__()`.","ref":"Oaskit.Spec.Paths.html#from_routes/2"},{"type":"function","title":"Oaskit.Spec.Paths.schema/0","doc":"","ref":"Oaskit.Spec.Paths.html#schema/0"},{"type":"module","title":"Oaskit.Spec.Reference","doc":"","ref":"Oaskit.Spec.Reference.html"},{"type":"function","title":"Oaskit.Spec.Reference.schema/0","doc":"","ref":"Oaskit.Spec.Reference.html#schema/0"},{"type":"module","title":"Oaskit.Spec.RequestBody","doc":"Representation of the [ Request Body Object](https://spec.openapis.org/oas/v3.1.1.html#request-body-object) in OpenAPI Specification.","ref":"Oaskit.Spec.RequestBody.html"},{"type":"function","title":"Oaskit.Spec.RequestBody.from_controller!/1","doc":"","ref":"Oaskit.Spec.RequestBody.html#from_controller!/1"},{"type":"function","title":"Oaskit.Spec.RequestBody.schema/0","doc":"","ref":"Oaskit.Spec.RequestBody.html#schema/0"},{"type":"module","title":"Oaskit.Spec.Response","doc":"Representation of the [ Response Object](https://spec.openapis.org/oas/v3.1.1.html#response-object) in OpenAPI Specification.","ref":"Oaskit.Spec.Response.html"},{"type":"function","title":"Oaskit.Spec.Response.from_controller!/1","doc":"","ref":"Oaskit.Spec.Response.html#from_controller!/1"},{"type":"function","title":"Oaskit.Spec.Response.schema/0","doc":"","ref":"Oaskit.Spec.Response.html#schema/0"},{"type":"module","title":"Oaskit.Spec.Responses","doc":"Representation of the [ Responses Object](https://spec.openapis.org/oas/v3.1.1.html#responses-object) in OpenAPI Specification.","ref":"Oaskit.Spec.Responses.html"},{"type":"function","title":"Oaskit.Spec.Responses.schema/0","doc":"","ref":"Oaskit.Spec.Responses.html#schema/0"},{"type":"module","title":"Oaskit.Spec.SchemaWrapper","doc":"","ref":"Oaskit.Spec.SchemaWrapper.html"},{"type":"function","title":"Oaskit.Spec.SchemaWrapper.schema/0","doc":"","ref":"Oaskit.Spec.SchemaWrapper.html#schema/0"},{"type":"module","title":"Oaskit.Spec.SecurityRequirement","doc":"Representation of the [ Security Requirement Object](https://spec.openapis.org/oas/v3.1.1.html#security-requirement-object) in OpenAPI Specification.","ref":"Oaskit.Spec.SecurityRequirement.html"},{"type":"function","title":"Oaskit.Spec.SecurityRequirement.schema/0","doc":"","ref":"Oaskit.Spec.SecurityRequirement.html#schema/0"},{"type":"module","title":"Oaskit.Spec.SecurityScheme","doc":"Representation of the [ Security Scheme Object](https://spec.openapis.org/oas/v3.1.1.html#security-scheme-object) in OpenAPI Specification.","ref":"Oaskit.Spec.SecurityScheme.html"},{"type":"function","title":"Oaskit.Spec.SecurityScheme.schema/0","doc":"","ref":"Oaskit.Spec.SecurityScheme.html#schema/0"},{"type":"module","title":"Oaskit.Spec.Server","doc":"Representation of the [ Server Object](https://spec.openapis.org/oas/v3.1.1.html#server-object) in OpenAPI Specification.","ref":"Oaskit.Spec.Server.html"},{"type":"function","title":"Oaskit.Spec.Server.from_config/2","doc":"Returns an URL based on the `:url` configuration of the endpoint.\n\nIf not all the parts of the URL are configured the following defaults are\nused:\n\n* `:port` defaults to `443`.\n* `:scheme` defaults to `\"https\"`.\n* `:path` defaults to `\"/\"`.","ref":"Oaskit.Spec.Server.html#from_config/2"},{"type":"function","title":"Oaskit.Spec.Server.schema/0","doc":"","ref":"Oaskit.Spec.Server.html#schema/0"},{"type":"module","title":"Oaskit.Spec.ServerVariable","doc":"Representation of the [ Server Variable Object](https://spec.openapis.org/oas/v3.1.1.html#server-variable-object) in OpenAPI Specification.","ref":"Oaskit.Spec.ServerVariable.html"},{"type":"function","title":"Oaskit.Spec.ServerVariable.schema/0","doc":"","ref":"Oaskit.Spec.ServerVariable.html#schema/0"},{"type":"module","title":"Oaskit.Spec.Tag","doc":"Representation of the [ Tag Object](https://spec.openapis.org/oas/v3.1.1.html#tag-object) in OpenAPI Specification.","ref":"Oaskit.Spec.Tag.html"},{"type":"function","title":"Oaskit.Spec.Tag.schema/0","doc":"","ref":"Oaskit.Spec.Tag.html#schema/0"},{"type":"module","title":"Oaskit.Spec.XML","doc":"Representation of the [ Xml Object](https://spec.openapis.org/oas/v3.1.1.html#xml-object) in OpenAPI Specification.","ref":"Oaskit.Spec.XML.html"},{"type":"function","title":"Oaskit.Spec.XML.schema/0","doc":"","ref":"Oaskit.Spec.XML.html#schema/0"},{"type":"module","title":"Oaskit.Parsers.HttpStructuredField","doc":"HTTP Structured Field parser implementation following RFC 8941.","ref":"Oaskit.Parsers.HttpStructuredField.html"},{"type":"function","title":"Oaskit.Parsers.HttpStructuredField.parse_sf_dictionary/2","doc":"","ref":"Oaskit.Parsers.HttpStructuredField.html#parse_sf_dictionary/2"},{"type":"function","title":"Oaskit.Parsers.HttpStructuredField.parse_sf_item/2","doc":"","ref":"Oaskit.Parsers.HttpStructuredField.html#parse_sf_item/2"},{"type":"function","title":"Oaskit.Parsers.HttpStructuredField.parse_sf_list/2","doc":"","ref":"Oaskit.Parsers.HttpStructuredField.html#parse_sf_list/2"},{"type":"function","title":"Oaskit.Parsers.HttpStructuredField.post_process_dict/2","doc":"","ref":"Oaskit.Parsers.HttpStructuredField.html#post_process_dict/2"},{"type":"function","title":"Oaskit.Parsers.HttpStructuredField.post_process_item/2","doc":"","ref":"Oaskit.Parsers.HttpStructuredField.html#post_process_item/2"},{"type":"function","title":"Oaskit.Parsers.HttpStructuredField.post_process_list/2","doc":"","ref":"Oaskit.Parsers.HttpStructuredField.html#post_process_list/2"},{"type":"module","title":"Oaskit.JsonSchema.Formats","doc":"Implements formats described in https://spec.openapis.org/api/format.json for\n`JSV`'s format validation. Some of the described formats are already\nimplemented in JSV directly.\n\nAll schemas used by this library can use those formats in any schema, on top\nof all formats already supported by `JSV`.\n\nUse the `c:Oaskit.jsv_opts/0` callback to add your own formats.\n\nThe `html` and `commonmark` formats are not validating anything and will\naccept any string.","ref":"Oaskit.JsonSchema.Formats.html"},{"type":"module","title":"Number formats - Oaskit.JsonSchema.Formats","doc":"Number formats apply to the `number` JSON type. So an `int16` format will be\nused on floats and reject them.\n\n`Decimal` values are supported.\n\n - `double`\n - `double-int`\n - `float`\n - `int16`\n - `int32`\n - `int8`\n - `uint16`\n - `uint32`\n - `uint8`","ref":"Oaskit.JsonSchema.Formats.html#module-number-formats"},{"type":"module","title":"Numeric formats - Oaskit.JsonSchema.Formats","doc":"Those apply to numbers and strings, as their value is not representable in\nevery programming language.\n\n`Decimal` values are supported.\n\n - `decimal`\n - `decimal128`\n - `int64`\n - `uint64`","ref":"Oaskit.JsonSchema.Formats.html#module-numeric-formats"},{"type":"module","title":"String formats - Oaskit.JsonSchema.Formats","doc":"String formats only apply when the value is a string, much like the default\nJSON schema formats like \"date\", \"uri\", _etc._.\n\n - `base64url`\n - `binary`\n - `byte`\n - `char`\n - `commonmark`\n - `html`\n - `media-range`\n - `password`","ref":"Oaskit.JsonSchema.Formats.html#module-string-formats"},{"type":"module","title":"HTTP Structured fields formats - Oaskit.JsonSchema.Formats","doc":"HTTP structured fields are generally used for headers. While this library does\nnot provide header validation yet, they are supported for OpenAPI\nspecification compliance.\n\nThe [`{:abnf_parsec, \"~> 2.0\"}`](https://hex.pm/packages/abnf_parsec)\ndependency is required for them to be available.\n\n - `sf-binary`\n - `sf-boolean`\n - `sf-decimal`\n - `sf-integer`\n - `sf-string`\n - `sf-token`","ref":"Oaskit.JsonSchema.Formats.html#module-http-structured-fields-formats"},{"type":"function","title":"Oaskit.JsonSchema.Formats.applies_to_type?/2","doc":"Callback implementation for  JSV.FormatValidator.applies_to_type?/2 .","ref":"Oaskit.JsonSchema.Formats.html#applies_to_type?/2"},{"type":"function","title":"Oaskit.JsonSchema.Formats.supported_formats/0","doc":"Callback implementation for  JSV.FormatValidator.supported_formats/0 .","ref":"Oaskit.JsonSchema.Formats.html#supported_formats/0"},{"type":"function","title":"Oaskit.JsonSchema.Formats.validate_cast/2","doc":"Callback implementation for  JSV.FormatValidator.validate_cast/2 .","ref":"Oaskit.JsonSchema.Formats.html#validate_cast/2"},{"type":"exception","title":"Oaskit.Errors.InvalidBodyError","doc":"","ref":"Oaskit.Errors.InvalidBodyError.html"},{"type":"function","title":"Oaskit.Errors.InvalidBodyError.message/1","doc":"Callback implementation for  Exception.message/1 .","ref":"Oaskit.Errors.InvalidBodyError.html#message/1"},{"type":"opaque","title":"Oaskit.Errors.InvalidBodyError.t/0","doc":"","ref":"Oaskit.Errors.InvalidBodyError.html#t:t/0"},{"type":"exception","title":"Oaskit.Errors.InvalidParameterError","doc":"","ref":"Oaskit.Errors.InvalidParameterError.html"},{"type":"function","title":"Oaskit.Errors.InvalidParameterError.message/1","doc":"Callback implementation for  Exception.message/1 .","ref":"Oaskit.Errors.InvalidParameterError.html#message/1"},{"type":"opaque","title":"Oaskit.Errors.InvalidParameterError.t/0","doc":"","ref":"Oaskit.Errors.InvalidParameterError.html#t:t/0"},{"type":"exception","title":"Oaskit.Errors.MissingParameterError","doc":"","ref":"Oaskit.Errors.MissingParameterError.html"},{"type":"function","title":"Oaskit.Errors.MissingParameterError.message/1","doc":"Callback implementation for  Exception.message/1 .","ref":"Oaskit.Errors.MissingParameterError.html#message/1"},{"type":"opaque","title":"Oaskit.Errors.MissingParameterError.t/0","doc":"","ref":"Oaskit.Errors.MissingParameterError.html#t:t/0"},{"type":"exception","title":"Oaskit.Errors.NormalizeError","doc":"","ref":"Oaskit.Errors.NormalizeError.html"},{"type":"function","title":"Oaskit.Errors.NormalizeError.message/1","doc":"Callback implementation for  Exception.message/1 .","ref":"Oaskit.Errors.NormalizeError.html#message/1"},{"type":"opaque","title":"Oaskit.Errors.NormalizeError.t/0","doc":"","ref":"Oaskit.Errors.NormalizeError.html#t:t/0"},{"type":"exception","title":"Oaskit.Errors.UnsupportedMediaTypeError","doc":"","ref":"Oaskit.Errors.UnsupportedMediaTypeError.html"},{"type":"function","title":"Oaskit.Errors.UnsupportedMediaTypeError.message/1","doc":"Callback implementation for  Exception.message/1 .","ref":"Oaskit.Errors.UnsupportedMediaTypeError.html#message/1"},{"type":"opaque","title":"Oaskit.Errors.UnsupportedMediaTypeError.t/0","doc":"","ref":"Oaskit.Errors.UnsupportedMediaTypeError.html#t:t/0"},{"type":"extras","title":"Changelog","doc":"# Changelog\n\nAll notable changes to this project will be documented in this file.\n\n## [0.2.0] - 2025-07-05","ref":"changelog.html"},{"type":"extras","title":"🚀 Features - Changelog","doc":"- Provide a JSON schema for the default error responses\n\n## [0.1.2] - 2025-06-30","ref":"changelog.html#features"},{"type":"extras","title":"🚀 Features - Changelog","doc":"- Provide a controller to serve the specs and Redoc UI\n\n## [0.1.1] - 2025-06-29","ref":"changelog.html#features-1"},{"type":"extras","title":"📚 Documentation - Changelog","doc":"- Fix doc rendering on hexdocs.pm","ref":"changelog.html#documentation"},{"type":"extras","title":"⚙️ Miscellaneous Tasks - Changelog","doc":"- Setup versioning\n\n## [0.1.0] - 2025-06-29","ref":"changelog.html#miscellaneous-tasks"},{"type":"extras","title":"🚀 Features - Changelog","doc":"- Initialize from proof of concept","ref":"changelog.html#features-2"},{"type":"extras","title":"⚙️ Miscellaneous Tasks - Changelog","doc":"- Hex package setup\n\n<!-- generated by git-cliff -->","ref":"changelog.html#miscellaneous-tasks-1"},{"type":"extras","title":"Oaskit","doc":"# Oaskit\n\n<!-- rdmx :badges\n    hexpm         : \"oaskit?color=4e2a8e\"\n    github_action : \"lud/oaskit/elixir.yaml?label=CI&branch=main\"\n    license       : oaskit\n    -->\n[![hex.pm Version](https://img.shields.io/hexpm/v/oaskit?color=4e2a8e)](https://hex.pm/packages/oaskit)\n[![Build Status](https://img.shields.io/github/actions/workflow/status/lud/oaskit/elixir.yaml?label=CI&branch=main)](https://github.com/lud/oaskit/actions/workflows/elixir.yaml?query=branch%3Amain)\n[![License](https://img.shields.io/hexpm/l/oaskit.svg)](https://hex.pm/packages/oaskit)\n<!-- rdmx /:badges -->\n\nOaskit is a set of macros and plugs for Elixir/Phoenix applications to\nautomatically validate incoming HTTP requests based on the [OpenAPI\nSpecification v3.1](https://spec.openapis.org/oas/v3.1.1.html).\n\n* Request bodies, path and query parameters validation with JSON schemas\n  supported by [JSV](https://hex.pm/packages/jsv).\n* Heavily inspired by [OpenApiSpex](https://hex.pm/packages/open_api_spex).\n* Mix task for JSON file specification generation.","ref":"readme.html"},{"type":"extras","title":"Documentation - Oaskit","doc":"The [Documentation](https://hexdocs.pm/oaskit/) is available on hexdocs,\nincluding a [Quickstart Guide](https://hexdocs.pm/oaskit/quickstart.html) to\ndive right in.","ref":"readme.html#documentation"},{"type":"extras","title":"Installation - Oaskit","doc":"<!-- rdmx :app_dep vsn:$app_vsn -->\n```elixir\ndef deps do\n  [\n    {:oaskit, \"~> 0.2\"},\n  ]\nend\n```\n<!-- rdmx /:app_dep -->\n\nYou can also import formatter rules in your `.formatter.exs` file:\n\n```elixir\n[\n  import_deps: [:oaskit]\n]\n```","ref":"readme.html#installation"},{"type":"extras","title":"Contributing - Oaskit","doc":"Pull requests are welcome, provided they include appropriate tests and\ndocumentation.","ref":"readme.html#contributing"},{"type":"extras","title":"Roadmap - Oaskit","doc":"* Serve SwaggerUI or redoc.\n* Provide header validation.\n* Define JSON schemas for the default error handler responses.","ref":"readme.html#roadmap"},{"type":"extras","title":"Oaskit Quickstart Guide","doc":"# Oaskit Quickstart Guide\n\nWelcome to Oaskit! This guide will walk you through setting up and using Oaskit to validate HTTP requests in your Phoenix application based on OpenAPI 3.1 specifications.","ref":"quickstart.html"},{"type":"extras","title":"Installation - Oaskit Quickstart Guide","doc":"First, add Oaskit to your dependencies in `mix.exs`:\n\n<!-- rdmx :app_dep vsn:$app_vsn -->\n```elixir\ndef deps do\n  [\n    {:oaskit, \"~> 0.1\"},\n  ]\nend\n```\n<!-- rdmx /:app_dep -->\n\nYou can also add the [AbnfParsec](https://hex.pm/packages/abnf_parsec)\ndependency to support [more\nformats](https://hexdocs.pm/jsv/validation-basics.html#formats) in the JSON\nschema `format` keyword.\n\nFinally, you will most probably use macros from this library, import formatter\nrules in your `.formatter.exs` file:\n\n<!-- rdmx :section name:formatter_config format: true -->\n```elixir\n# .formatter.exs\n[\n  import_deps: [:oaskit]\n]\n```\n<!-- rdmx /:section -->","ref":"quickstart.html#installation"},{"type":"extras","title":"Creating an API Specification Module - Oaskit Quickstart Guide","doc":"Create a module that defines your OpenAPI specification using the `Oaskit`\nmodule. This module will serve as the central definition of your API:\n\n<!-- rdmx :section name:api_spec_module format: true -->\n```elixir\ndefmodule MyAppWeb.ApiSpec do\n  alias Oaskit.Spec.Paths\n  alias Oaskit.Spec.Server\n  use Oaskit\n\n  @impl true\n  def spec do\n    %{\n      openapi: \"3.1.1\",\n      info: %{\n        title: \"My App API\",\n        version: \"1.0.0\",\n        description: \"Main HTTP API for My App\"\n      },\n      servers: [Server.from_config(:my_app, MyAppWeb.Endpoint)],\n      paths: Paths.from_router(MyAppWeb.Router, filter: &String.starts_with?(&1.path, \"/api/\"))\n    }\n  end\nend\n```\n<!-- rdmx /:section -->\n\nThe `Oaskit.Spec.Paths.from_router/2` function automatically extracts API\npaths from your Phoenix router, focusing only on controller actions that have\noperations defined. The optional `:filter` function lets you limit which routes\nare included in your specification.","ref":"quickstart.html#creating-an-api-specification-module"},{"type":"extras","title":"Setting Up Router Pipelines - Oaskit Quickstart Guide","doc":"Configure your Phoenix router to use Oaskit validation with your API spec\nmodule by setting up a pipeline with the `Oaskit.Plugs.SpecProvider` plug:\n\n<!-- rdmx :section name:router_pipeline format: true -->\n```elixir\ndefmodule MyAppWeb.Router do\n  use Phoenix.Router\n\n  # Define a pipeline for API routes with Oaskit validation\n  pipeline :api do\n    plug :accepts, [\"json\"]\n    plug Oaskit.Plugs.SpecProvider, spec: MyAppWeb.ApiSpec\n  end\n\n  # Apply the pipeline to your API routes\n  scope \"/api\", MyAppWeb do\n    pipe_through :api\n\n    resources \"/users\", UserController, only: [:index, :show, :create]\n    resources \"/posts\", PostController, only: [:index, :show, :create, :update]\n  end\nend\n```\n<!-- rdmx /:section -->\n\nThe `Oaskit.Plugs.SpecProvider` plug is required as a controller action and\nits defined OpenAPI operation can be referenced in multiple specifications.","ref":"quickstart.html#setting-up-router-pipelines"},{"type":"extras","title":"Configuring Controllers - Oaskit Quickstart Guide","doc":"Set up your controllers to use Oaskit validation. You can do this globally in\nyour `MyAppWeb` module:\n\n<!-- rdmx :section name:controller_config format: true -->\n```elixir\ndefmodule MyAppWeb do\n  def controller do\n    quote do\n      use Phoenix.Controller,\n        formats: [:html, :json],\n        layouts: [html: MyAppWeb.Layouts]\n\n      # Add Oaskit controller macros\n      use Oaskit.Controller\n\n      # Add the validation plug\n      plug Oaskit.Plugs.ValidateRequest\n\n      import Plug.Conn\n      use Gettext, backend: MyAppWeb.Gettext\n\n      unquote(verified_routes())\n    end\n  end\n\n  # ...\nend\n```\n<!-- rdmx /:section -->\n\nThis setup ensures all controllers using `use MyAppWeb, :controller` will have\nOaskit validation enabled. See the documentation of\n`Oaskit.Plugs.ValidateRequest` to configure the plug globally for only a\nsubset of your controllers.","ref":"quickstart.html#configuring-controllers"},{"type":"extras","title":"Defining Operations in Controllers - Oaskit Quickstart Guide","doc":"Now you can define operations in your controllers using the\n`Oaskit.Controller.operation/2` macro.\n\nTo import the macros, use your `:controller` helper as usual since you've added\n`use Oaskit.Controller` in the previous step.\n\n<!-- rdmx :section name:controller_use format: true -->\n```elixir\ndefmodule MyAppWeb.UserController do\n  use MyAppWeb, :controller\n\n  # ...\nend\n```\n<!-- rdmx /:section -->\n\nThe operation macro takes the function name and the operation specs as\narguments.\n\nIn this example we use syntax shortcuts to refer directly to schemas, but the\noperation macro lets you define responses or requests with multiple content\ntypes, custom descriptions and many other options. Make sure to read the docs!\n\n<!-- rdmx :section name:operation_definition format: true -->\n```elixir\noperation :create,\n  summary: \"Create a new user\",\n  request_body: {\n    %{\n      type: :object,\n      properties: %{\n        name: %{type: :string, minLength: 1},\n        email: %{type: :string, format: :email},\n        age: %{type: :integer, minimum: 18}\n      },\n      required: [:name, :email]\n    },\n    description: \"The user payload\"\n  },\n  parameters: [\n    source: [in: :query, schema: %{type: :string}, required: false]\n  ],\n  responses: [\n    created: UserSchema,\n    unprocessable_entity: ErrorSchema\n  ]\n\ndef create(conn, _params) do\n  # The `_params` variable from phoenix is not changed by the validation plug.\n  #\n  # Validated and cast data is stored in `conn.private.oaskit`.\n  #\n  # You may explore what's in there (or read the docs), or you may use the\n  # various helpers from the `Oaskit.Controller` module:\n  user_data = body_params(conn)\n  source = query_param(conn, :source)\n\n  case create_user(user_data, source) do\n    # ...\n  end\nend\n```\n<!-- rdmx /:section -->","ref":"quickstart.html#defining-operations-in-controllers"},{"type":"extras","title":"Defining JSON Schemas - Oaskit Quickstart Guide","doc":"There are two ways to provide schemas in the various macros, either inline or\nusing modules.\n\n#### Inline schemas\n\nInline schemas are maps (with atoms or binary keys and values) or booleans.\n\n<!-- rdmx :section name:inline_schemas format: true -->\n```elixir\n@user_schema %{\n  type: :object,\n  title: \"User\",\n  properties: %{\n    name: %{type: :string, minLength: 1},\n    email: %{type: :string, format: :email},\n    age: %{type: :integer, minimum: 0}\n  },\n  required: [:name, :email]\n}\n\noperation :create,\n  request_body: {@user_schema, [required: true]},\n  responses: [ok: {@user_schema, []}]\n```\n<!-- rdmx /:section -->\n\nWhile they are practical, such maps are duplicated in the compiled module as\nwell as in the OpenAPI specification document.\n\n\n#### Module-based schemas\n\nA module-based schema is any module that exports a `schema/0` function returning\na valid JSON schema.\n\nOaskit uses [JSV](https://hex.pm/packages/jsv). Module-based schemas defined\nwith `JSV.defschema/1` are automatically cast to structs when validation\nsucceeds, making them convenient to work with, notably thanks to the new Elixir\ntypes compiler!\n\nMake sure to check the [JSV\ndocumentation](https://hexdocs.pm/jsv/defining-schemas.html) for additional\nfeatures.\n\n<!-- rdmx :section name:module_schemas format: true -->\n```elixir\ndefmodule MyAppWeb.Schemas.UserSchema do\n  use JSV.Schema\n\n  defschema %{\n    type: :object,\n    title: \"User\",\n    properties: %{\n      id: %{type: :integer},\n      name: %{type: :string, minLength: 1},\n      email: %{type: :string, format: :email},\n      age: %{type: :integer, minimum: 18}\n    },\n    required: [:id, :name, :email]\n  }\nend\n\n# Use the module directly in operations in place of a schema\noperation :show,\n  responses: [ok: MyAppWeb.Schemas.UserSchema]\n```\n<!-- rdmx /:section -->\n\nSuch schemas are collected into the `#/components/schemas` section of the\nOpenAPI specification, which makes that document shorter, easier to navigate,\nand limits the memory usage at runtime.","ref":"quickstart.html#defining-json-schemas"},{"type":"extras","title":"Shared tags and parameters - Oaskit Quickstart Guide","doc":"Oaskit provides the `Oaskit.Controller.tags/1` and\n`Oaskit.Controller.parameter/2` macros for shared elements between\noperations.\n\nThose macros only apply to operations defined _after_ them.\n\n<!-- rdmx :section name:shared_elements format: true -->\n```elixir\n# Add tags to group operations in documentation\ntags [\"users\", \"public\"]\n\n# Define parameters once for multiple operations\nparameter :page, in: :query, schema: %{type: :integer}\nparameter :per_page, in: :query, schema: %{type: :integer}\n```\n<!-- rdmx /:section -->","ref":"quickstart.html#shared-tags-and-parameters"},{"type":"extras","title":"Testing with Oaskit.Test - Oaskit Quickstart Guide","doc":"The `valid_response/3` helper validates that the response matches your OpenAPI\nspecification, including status code, content type, and response body schema. It\nreturns the parsed response data for further assertions.\n\n<!-- rdmx :section name:test_example format: true -->\n```elixir\ndefmodule MyAppWeb.UserControllerTest do\n  use MyAppWeb.ConnCase\n\n  # Helper to wrap Oaskit.Test.valid_response/3\n  # Feel free to add it directly into your ConnCase module!\n  defp valid_response(conn, status) do\n    Oaskit.Test.valid_response(MyAppWeb.ApiSpec, conn, status)\n  end\n\n  test \"create user with valid data\", %{conn: conn} do\n    user_params = %{\n      name: \"John Doe\",\n      email: \"john@example.com\",\n      age: 25\n    }\n\n    conn = post(conn, ~p\"/api/users\", user_params)\n\n    # Validate the response against your OpenAPI specification. It returns\n    # decoded data for JSON content-types.\n    assert %{\n             \"name\" => \"John Doe\",\n             \"email\" => \"john@example.com\",\n             \"age\" => 25\n           } =\n             valid_response(conn, 201)\n  end\n\n  test \"create user with invalid data returns validation errors\", %{conn: conn} do\n    invalid_params = %{\n      name: \"\",\n      email: \"invalid-email\",\n      age: 15\n    }\n\n    conn = post(conn, ~p\"/api/users\", invalid_params)\n\n    # You can use `valid_response` if you define a response schema for the\n    # errors. See `Oaskit.ErrorHandler.Default.error_response_schema/0`.\n    #\n    # valid_response(conn, 422)\n\n    # If you do not declare all possible responses, using the good old\n    # `Phoenix.ConnTest.json_response/2` works fine!\n    assert json_response(conn, 422)\n  end\nend\n```\n<!-- rdmx /:section -->","ref":"quickstart.html#testing-with-oaskit-test"},{"type":"extras","title":"Generating OpenAPI Documentation - Oaskit Quickstart Guide","doc":"Once you have your operations defined, you can generate an OpenAPI specification\nfile using the Mix task:\n\n<!-- rdmx :section name:openapi_dump format: true -->\n```bash\nmix openapi.dump MyAppWeb.ApiSpec --pretty -o priv/openapi.json\n```\n<!-- rdmx /:section -->\n\nThis generates a complete OpenAPI 3.1 specification file that can be used with\nvarious tools like client generators for TypeScript or Elixir.","ref":"quickstart.html#generating-openapi-documentation"},{"type":"extras","title":"Serving OpenAPI Specifications - Oaskit Quickstart Guide","doc":"You can serve your OpenAPI specification dynamically using the\n`Oaskit.SpecController`. This controller serves your specification as JSON at an\nHTTP endpoint.\n\nAdd the controller to your router:\n\n<!-- rdmx :section name:spec_controller format: true -->\n```elixir\nget \"/openapi.json\", Oaskit.SpecController, spec: MyAppWeb.ApiSpec\n```\n<!-- rdmx /:section -->\n\nThis will serve your OpenAPI specification at `/openapi.json`. Pass `?pretty=1`\nto get pretty printed JSON.\n\nYou can also use that controller to serve Redoc UI. Pass the full URL path to\nthe json route you just defined:\n\n<!-- rdmx :section name:redoc_controller format: true -->\n```elixir\nget \"/docs\", Oaskit.SpecController, redoc: \"/openapi.json\"\n```\n<!-- rdmx /:section -->\n\nRedoc allows you to browse your API endpoints, view request/response schemas,\nand see examples, but note that it is read-only and doesn't allow testing the\nAPI directly.","ref":"quickstart.html#serving-openapi-specifications"},{"type":"extras","title":"Ask for help! - Oaskit Quickstart Guide","doc":"If anything is unclear, or if you would like to see more features, plase fill an\nissue in the [Github repository](https://github.com/lud/oaskit).\n\nHappy coding!","ref":"quickstart.html#ask-for-help"},{"type":"extras","title":"Using an External Specification","doc":"# Using an External Specification\n\nThis guide shows how to use Oaskit with existing OpenAPI specification documents\n(JSON or YAML files) to validate HTTP requests and responses in your Phoenix\napplication.","ref":"external-specs.html"},{"type":"extras","title":"Overview - Using an External Specification","doc":"While Oaskit can generate OpenAPI specifications from your controllers using the\n`operation/2` macro, you may want to work with existing specification documents\ninstead. This approach is useful when:\n\n* You have an existing OpenAPI specification that you need to implement.\n* You prefer to write your API specification first, maybe in cooperation with\n  frontend developers or consumers of the API.\n* You are generating the specification by other means.","ref":"external-specs.html#overview"},{"type":"extras","title":"Spec Module Setup - Using an External Specification","doc":"Create a module that implements the `Oaskit` behavior and returns your OpenAPI\nspecification.\n\nThen read the spec from the external source and return it from the `spec/0`\ncallback. It's better to read the external source at compile time if possible,\nto avoid unnecessary resource consumption. You can store the spec in an\nattribute during compilation.\n\nBelow is an example with JSON. Oaskit does not ship with a YAML parsing library,\nbut you can of course bring your own. As the `c:Oaskit.spec/0` callback must\nreturn data, and not a raw JSON string, you can actually parse that data from any\nsource.\n\n<!-- rdmx :section name:external_json_spec format: true -->\n```elixir\ndefmodule MyAppWeb.ExternalApiSpec do\n  use Oaskit\n\n  @moduledoc \"\"\"\n  OpenAPI specification loaded from an external document.\n  \"\"\"\n\n  # Load from a JSON file\n  @api_spec \"priv/openapi/my-api.json\"\n            |> File.read!()\n            |> JSON.decode!()\n\n  @impl true\n  def spec, do: @api_spec\nend\n```\n<!-- rdmx /:section -->\n\nYou can also define the specification directly in Elixir:\n\n<!-- rdmx :section name:manual_spec format: true -->\n```elixir\ndefmodule MyAppWeb.ManualApiSpec do\n  use Oaskit\n\n  @api_spec %{\n    \"openapi\" => \"3.1.1\",\n    \"info\" => %{\n      \"title\" => \"My API\",\n      \"version\" => \"1.0.0\"\n    },\n    \"paths\" => %{\n      \"/users\" => %{\n        \"get\" => some_operation\n      }\n    }\n  }\n\n  @impl true\n  def spec, do: @api_spec\nend\n```\n<!-- rdmx /:section -->","ref":"external-specs.html#spec-module-setup"},{"type":"extras","title":"Router Configuration - Using an External Specification","doc":"At the router level, there is no difference between generated specs and\n\"external\" specs, as Oaskit will build validation from the full normalized spec.\nJust use the `Oaskit.Plugs.SpecProvider` plug as for any spec module.\n\n<!-- rdmx :section name:external_router_config format: true -->\n```elixir\ndefmodule MyAppWeb.Router do\n  use Phoenix.Router\n\n  pipeline :api do\n    plug :accepts, [\"json\"]\n    plug Oaskit.Plugs.SpecProvider, spec: MyAppWeb.ExternalApiSpec\n  end\n\n  scope \"/api\", MyAppWeb do\n    pipe_through :api\n\n    resources \"/users\", UserController, only: [:index, :create]\n  end\nend\n```\n<!-- rdmx /:section -->","ref":"external-specs.html#router-configuration"},{"type":"extras","title":"Using Operations in Controllers - Using an External Specification","doc":"Instead of defining operations with the `operation/2` macro, call the\n`use_operation/2` macro to attach operations from your specification to action\nfunctions by their `operationId`.\n\nThis will instruct the `Oaskit.Plugs.ValidateRequest` plug to use the validations\ndefined in the spec when that action is called.\n\n<!-- rdmx :section name:use_operation_macro format: true -->\n```elixir\ndefmodule MyAppWeb.UserController do\n  use MyAppWeb, :controller\n\n  use_operation :index, \"listUsers\"\n\n  def index(conn, _params) do\n    # ...\n  end\n\n  use_operation :create, \"createUser\"\n\n  def create(conn, _params) do\n    # ...\n  end\nend\n```\n<!-- rdmx /:section -->","ref":"external-specs.html#using-operations-in-controllers"},{"type":"extras","title":"Controller Configuration - Using an External Specification","doc":"Ensure your controllers are configured to use Oaskit validation, typically in\nyour `MyAppWeb` module.\n\nThis is the same setup as for the classic usage of the library with the `operation`\nmacro.\n\n* Use `Oaskit.Controller` to pull the macros in and inject the setup code\n* Plug `Oaskit.Plugs.ValidateRequest` to validate everything when a request hits\n  the controller\n\n<!-- rdmx :section name:external_controller_config format: true -->\n```elixir\ndefmodule MyAppWeb do\n  def controller do\n    quote do\n      use Phoenix.Controller,\n        formats: [:html, :json],\n        layouts: [html: MyAppWeb.Layouts]\n\n      # Add Oaskit controller macros\n      use Oaskit.Controller\n\n      # Add the validation plug\n      plug Oaskit.Plugs.ValidateRequest\n\n      import Plug.Conn\n      use Gettext, backend: MyAppWeb.Gettext\n\n      unquote(verified_routes())\n    end\n  end\n\n  # ...\nend\n```\n<!-- rdmx /:section -->","ref":"external-specs.html#controller-configuration"},{"type":"extras","title":"Working with HTTP Methods - Using an External Specification","doc":"If your action function handles multiple operations, specify the method\nexplicitly:\n\n<!-- rdmx :section name:multiple_operations format: true -->\n```elixir\ndefmodule MyAppWeb.UserController do\n  use MyAppWeb, :controller\n\n  use_operation :generic_handler, \"listUsers\", method: :get\n  use_operation :generic_handler, \"createUser\", method: :post\n\n  def generic_handler(conn, _params) do\n    # ...\n  end\nend\n```\n<!-- rdmx /:section -->","ref":"external-specs.html#working-with-http-methods"},{"type":"extras","title":"Parameter Handling - Using an External Specification","doc":"> #### Parameter names always create atoms {: .warning}\n>\n> Query and path parameters defined in OpenAPI specifications always define\n> the corresponding atoms, even if that specification is read from a JSON\n> file, or defined manually in code with string keys.\n>\n> For that reason it is ill advised to use specs generated dynamically at\n> runtime without validating their content.\n\nThe controller helpers to get path params, query params, and body params are\nstill working, and everything is stored under `conn.private.oaskit`.\n\n<!-- rdmx :section name:external_parameter_handling format: true -->\n```elixir\ndef index(conn, _params) do\n  # Path parameters\n  slug = path_param(conn, :id)\n\n  # Query parameters\n  # defaults to 1\n  page = query_param(conn, :page, 1)\n  # defaults to nil\n  per_page = query_param(conn, :per_page)\n\n  # Request body (for POST/PUT/PATCH requests)\n  user_data = body_params(conn)\n\n  # ...\nend\n```\n<!-- rdmx /:section -->","ref":"external-specs.html#parameter-handling"},{"type":"extras","title":"Test with ExUnit - Using an External Specification","doc":"The `Oaskit.Test.valid_response/3` helper works seamlessly with external\nspecifications:\n\n<!-- rdmx :section name:external_testing format: true -->\n```elixir\ndefmodule MyAppWeb.UserControllerTest do\n  use MyAppWeb.ConnCase\n\n  # Helper to wrap Oaskit.Test.valid_response/3.\n  # Define it in your ConnCase module.\n  def valid_response(conn, status) do\n    Oaskit.Test.valid_response(MyAppWeb.ExternalApiSpec, conn, status)\n  end\n\n  test \"list users returns valid response\", %{conn: conn} do\n    conn = get(conn, ~p\"/api/users?page=1&per_page=5\")\n    assert %{\"users\" => users} = valid_response(conn, 200)\n\n    # ...\n  end\n\n  test \"create user with valid data\", %{conn: conn} do\n    user_params = %{\n      name: \"John Doe\",\n      email: \"john@example.com\"\n    }\n\n    conn = post(conn, ~p\"/api/users\", user_params)\n\n    # Validate against specification and get the JSON data\n    assert %{\n             \"id\" => id,\n             \"name\" => \"John Doe\",\n             \"email\" => \"john@example.com\"\n           } = valid_response(conn, 201)\n\n    # ...\n  end\n\n  test \"create user with invalid data returns validation errors\", %{conn: conn} do\n    invalid_params = %{\n      name: \"\",\n      email: \"invalid-email\"\n    }\n\n    conn = post(conn, ~p\"/api/users\", invalid_params)\n\n    # You can use `valid_response` if you define a response schema for the\n    # errors. See `Oaskit.ErrorHandler.Default.error_response_schema/0`.\n    #\n    # valid_response(conn, 422)\n\n    # If you do not declare all possible responses, using the good old\n    # `Phoenix.ConnTest.json_response/2` works fine!\n    assert json_response(conn, 422)\n  end\nend\n```\n<!-- rdmx /:section -->","ref":"external-specs.html#test-with-exunit"},{"type":"extras","title":"Serving the Specification - Using an External Specification","doc":"You can serve your external specification using the same `Oaskit.SpecController`\nas with generated specs:\n\n<!-- rdmx :section name:external_spec_serving format: true -->\n```elixir\n# In your router\nscope \"/\" do\n  pipe_through :api\n  get \"/openapi.json\", Oaskit.SpecController, :show\n  get \"/docs\", Oaskit.SpecController, redoc: \"/openapi.json\"\nend\n```\n<!-- rdmx /:section -->\n\nBut remember that if your external spec is a static JSON file, you can just\nserve that file with `Plug.Static` or Nginx, Apache, etc.","ref":"external-specs.html#serving-the-specification"},{"type":"extras","title":"Mixing Approaches - Using an External Specification","doc":"Finally, you can combine both approaches in the same application. This is useful\nwhen you want to implement a predefined API but also want to add new routes.\n\n<!-- rdmx :section name:hybrid_spec format: true -->\n```elixir\ndefmodule MyAppWeb.HybridApiSpec do\n  alias Oaskit.Spec.Paths\n  use Oaskit\n\n  @base_spec \"priv/openapi/core-api.json\"\n             |> File.read!()\n             |> Jason.decode!()\n\n  @impl true\n  def spec do\n    Map.update!(@base_spec, \"paths\", fn existing_paths ->\n      Map.merge(existing_paths, admin_paths())\n    end)\n  end\n\n  defp admin_paths do\n    Paths.from_router(\n      MyAppWeb.Router,\n      filter: &String.starts_with?(&1.path, \"/admin\")\n    )\n  end\nend\n```\n<!-- rdmx /:section -->\n\nYou can also merge different routers, paths from multiple files, _etc._ This will\nmerge together specs defined with atom keys and structs with specs defined only\nwith strings. It's not a problem since Oaskit will normalize it before building\nthe validation.\n\nThe only concern is to make sure that you do not override paths unknowingly if\nfor some reason one of the specs uses atoms for path keys.","ref":"external-specs.html#mixing-approaches"}],"proglang":"elixir","content_type":"text/markdown","producer":{"name":"ex_doc","version":"0.38.2"}}